import time
from typing import Any, Dict, Optional
from .utils import are_object_equals
from .result import OrquestaResult


class OrquestaCacheItem:
    def __init__(
        self, key: str, result: OrquestaResult, context: Dict[str, Any], created: int
    ):
        self.key = key
        self.result = result
        self.context = context
        self.created = created


class CacheStore:
    def __init__(self, ttl: int):
        self.cache: list[OrquestaCacheItem] = []
        self.ttl = ttl

    def get(self, key: str, context: Dict[str, Any]) -> Optional[OrquestaCacheItem]:
        return next(
            (
                item
                for item in self.cache
                if item.key == key
                and are_object_equals(item.context, context)
                and not self.is_expired(item.created, self.ttl)
            ),
            None,
        )

    def set(self, key: str, result: OrquestaResult, context: Dict[str, Any]):
        self.cache.append(
            OrquestaCacheItem(key, result, context, int(time.time() * 1000))
        )

    def is_expired(self, cacheTime: int, ttl: int) -> bool:
        return time.time() * 1000 - cacheTime > ttl * 1000

    def flush_expired(self):
        self.cache = [
            item for item in self.cache if not self.is_expired(item.created, self.ttl)
        ]



def is_invalid_api_key(key):
    return not key.startswith("RQST.")


def are_object_equals(dict1, dict2):
    if isinstance(dict1, dict) and isinstance(dict2, dict):
        keys1 = sorted(dict1.keys())
        keys2 = sorted(dict2.keys())

        if keys1 != keys2:
            return False

        for key in keys1:
            value1 = dict1[key]
            value2 = dict2[key]

            if isinstance(value1, dict) and isinstance(value2, dict):
                if not are_object_equals(value1, value2):
                    return False
            elif value1 != value2:
                return False

        return True

    return False



import time
from typing import Any, Dict, Optional

from .cache import CacheStore
from .request import RULES_URL
from .result import OrquestaBaseEntity, OrquestaResult
from .rule_type import OrquestaRuleType


class OrquestaRulesMetrics:
    def __init__(self, metadata: Dict[str, Any]):
        self.metadata = metadata


class OrquestaRuleResult(OrquestaResult):
    def __init__(
        self,
        dsn: str,
        start_time: int,
        value: Any,
        trace_id: Optional[str],
        type: Optional[OrquestaRuleType],
    ):
        super().__init__(dsn, RULES_URL, start_time, value, trace_id, type)
        self.default_matched = not trace_id


class OrquestaRules(OrquestaBaseEntity):
    def __init__(self, dsn: str, cache: CacheStore):
        super().__init__(dsn, RULES_URL, "rules")

        self.__cache = cache

    def query(
        self,
        rule_key: str,
        default_value: Any,
        context: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> OrquestaRuleResult:
        cached_result = self.__cache.get(rule_key, context or {})

        if cached_result:
            return cached_result.result

        start_time = int(time.time() * 1000)

        response = self.perform_request(rule_key, context, metadata)

        result = response.json().get(rule_key, {})

        rule_result = OrquestaRuleResult(
            self.dsn,
            start_time,
            result.get("value", default_value),
            result.get("trace_id"),
            result.get("type"),
        )

        if result.get("trace_id") and result.get("value") is not None:
            self.__cache.set(rule_key, rule_result, context or {})

        return rule_result



import time
from typing import Any, Dict, Optional

from .cache import CacheStore
from .request import PROMPTS_URL
from .result import OrquestaBaseEntity, OrquestaResult
from .rule_type import OrquestaRuleType


class OrquestaPromptMetricsEconomics:
    def __init__(self, prompt_tokens: int, completion_tokens: int, total_tokens: int):
        self.prompt_tokens = prompt_tokens
        self.completion_tokens = completion_tokens
        self.total_tokens = total_tokens


class OrquestaPromptMetrics:
    def __init__(
        self,
        score: Optional[float] = None,
        latency: Optional[int] = None,
        metadata: Optional[Dict[str, Any]] = None,
        economics: Optional[OrquestaPromptMetricsEconomics] = None,
    ):
        self.score = score
        self.latency = latency
        self.metadata = metadata
        self.economics = economics


class OrquestaPromptResponse:
    def __init__(self, trace_id: str, value: Dict[str, Any], status: int):
        self.trace_id = trace_id
        self.value = value
        self.status = status


class OrquestaPromptResult(OrquestaResult):
    def __init__(
        self,
        api_key: str,
        start_time: int,
        value: Any,
        trace_id: Optional[str],
        type: Optional[OrquestaRuleType],
    ):
        super().__init__(api_key, PROMPTS_URL, start_time, value, trace_id, type)

        self.has_error = not bool(trace_id)


class OrquestaPrompts(OrquestaBaseEntity):
    def __init__(self, dsn: str, cache: CacheStore):
        super().__init__(dsn, PROMPTS_URL, "prompts")

        self.__cache = cache

    def query(
        self,
        prompt_key: str,
        context: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> OrquestaPromptResult:
        cached_result = self.__cache.get(prompt_key, context or {})

        if cached_result:
            return cached_result.result

        start_time = int(time.time() * 1000)

        response = self.perform_request(prompt_key, context, metadata)

        result = response.json().get(prompt_key, {})

        prompt_result = OrquestaPromptResult(
            self.dsn,
            start_time,
            result.get("value"),
            result.get("trace_id"),
            result.get("type"),
        )

        if result.get("trace_id") and result.get("value") is not None:
            self.__cache.set(prompt_key, prompt_result, context or {})

        return prompt_result



import time
from typing import Any, Optional

from .request import post
from .rule_type import OrquestaRuleType


class OrquestaResult:
    def __init__(
        self,
        api_key: str,
        endpoint_url: str,
        start_time: int,
        value: Any,
        trace_id: Optional[str],
        type: Optional[OrquestaRuleType],
    ):
        self.trace_id = trace_id
        self.type = type
        self.value = value
        self.__api_key = api_key
        self.__endpoint_url = endpoint_url

        self.__report_roundtrip_time(start_time)

    @property
    def _metrics_endpoint_url(self) -> str:
        return f"{self.__endpoint_url}/{self.trace_id}/metrics"

    def add_metrics(self, metrics: dict = {}) -> None:
        if not self.trace_id:
            return

        post(self._metrics_endpoint_url, self.__api_key, metrics)

    def __report_roundtrip_time(self, start_time: int) -> None:
        if not self.trace_id:
            return

        roundtrip_time = int(time.time() * 1000) - start_time

        post(
            self._metrics_endpoint_url,
            self.__api_key,
            {"roundtrip_time": roundtrip_time},
        )


class OrquestaBaseEntity:
    def __init__(
        self,
        dsn: str,
        endpointUrl: str,
        payloadKey: str,
    ):
        self.dsn = dsn
        self.__endpointUrl = endpointUrl
        self.__payloadKey = payloadKey

    def perform_request(
        self,
        key: str,
        context: Optional[dict] = None,
        metadata: Optional[dict] = None,
    ) -> Any:
        return post(
            self.__endpointUrl,
            self.dsn,
            {self.__payloadKey: [key], **(context or {}), **(metadata or {})},
        )



import requests
from .version import __version__

API_URL = "https://api.orquesta.dev/v2"
PROMPTS_URL = f"{API_URL}/prompts"
RULES_URL = f"{API_URL}/rules"


def post(url: str, api_key: str, body: dict):
    headers = {
        "X-SDK-Version": "orquestadev/{};python".format(__version__),
        "Authorization": f"Bearer {api_key}",
        "Content-Type": "application/json",
        "Accept": "application/json",
        "User-Agent": "orquestadev/{};python".format(__version__),
    }

    response = requests.post(url, json=body, headers=headers)
    return response



import os

from .rules import OrquestaRules
from .prompts import OrquestaPrompts

from .cache import CacheStore
from .utils import is_invalid_api_key


class OrquestaIncorrectAPIException(BaseException):
    """Raised if the provider API key is invalid."""

    pass


class OrquestaClienOptions:
    def __init__(self, ttl: int) -> None:
        self.ttl = ttl


class OrquestaClient:
    def __init__(self, api_key: str, options: OrquestaClienOptions = None):
        ttl = options.ttl if options else 3600

        cache = CacheStore(ttl=ttl)

        api_key = api_key or os.environ.get("ORQUESTA_API_KEY")

        if not api_key or is_invalid_api_key(api_key):
            raise OrquestaIncorrectAPIException("The provided API key is invalid.")

        self.rules = OrquestaRules(api_key, cache)
        self.prompts = OrquestaPrompts(api_key, cache)
